using System;
using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Vinput
{
    public class AutoGensThruster : MVRScript
    {
        // ============================================================================================== //
        // ========================================= PROPERTIES ========================================= //
        // ============================================================================================== //
        public bool __DEBUG__ = false;

        // UI variables
        private JSONStorableBool _activeJSON;
        private JSONStorableFloat _distanceInJSON;
        private JSONStorableFloat _distanceOutJSON;
        private JSONStorableFloat _distanceRandomJSON;
        private JSONStorableBool _thrustJSON;
        private JSONStorableBool _alignJSON;
        private JSONStorableBool _lockJSON;
        private JSONStorableFloat _dirOffsetXJSON;
        private JSONStorableFloat _dirOffsetYJSON;
        private JSONStorableStringChooser _easingInJSON;
        private JSONStorableStringChooser _easingOutJSON;
        private JSONStorableBool _useMasterJSON;

        // Person atoms, controls, colliders
        private Cycler _cycler;
        private JSONStorableFloat _cyclePctInJSON;

        private TargetManager _targetManager;
        private GensAligner _aligner;
        private FreeControllerV3 _penisBaseCtrl;

        // Temporary controller save/restore settings
        private CtrlSnapshot _ctrlSnapshot;
        private bool _restoring = false;
        private float _distanceFactor = 1.0f;

        // Getters
        private bool _isValid => _targetManager.TargetRB1 != null && _targetManager.TargetRB2 != null && _penisBaseCtrl != null;
        private float _thrustDistance => _distanceInJSON.val - _distanceOutJSON.val;

        private SyncManager _sync;
        public JSONStorableBool isActive = new JSONStorableBool("isActive", false);
        public JSONStorableFloat cyclePct = new JSONStorableFloat("cyclePct", 0f, 0f, 1f);
        private float _lastPct = 0f;


        // ============================================================================================== //
        // ========================================== METHODS =========================================== //
        // ============================================================================================== //

        // ========================================== BUILT-IN ========================================== //
        // Called by game engine on initialization
        public override void Init()
        {
            _targetManager = new TargetManager()
            {
                OnTargetChange = _ => { _UpdateState(); }
            };
            _aligner = new GensAligner(containingAtom, _targetManager);


            try
            {
                _cycler = new Cycler(this, OnThrustBegin, OnThrustUpdate);
                _sync = new SyncManager(this);
                RegisterBool(isActive);
                RegisterFloat(cyclePct);
                SetupUI();
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(AutoGensThruster)}.{nameof(Init)}: {e}");
            }

        }

        // Called by game engine before first update
        private void Start()
        {
            try
            {
                _penisBaseCtrl = containingAtom.freeControllers.First(fc => fc.name == "penisBaseControl");
            }
            catch (Exception e)
            {
                SuperController.LogError($"{nameof(AutoGensThruster)}: Plugin cannot be applied to an Atom without penisBaseControl.");
                return;
            }

            _UpdateState();

        }

        public void OnEnable() { _UpdateState(); }
        public void OnDisable() { _UpdateState(); }
        public void OnDestroy() { _Deactivate(); }


        // ========================================== CUSTOM ========================================== //

        private void OnThrustBegin()
        {
            if (_useMasterJSON.val) return;
            _distanceFactor = UnityEngine.Random.Range(1 - _distanceRandomJSON.val, 1 + _distanceRandomJSON.val);
        }

        private void OnThrustUpdate(float pct)
        {
            cyclePct.val = pct;
            if (_useMasterJSON.val)
            {
                if (_sync.masterActive)
                {
                    pct = _sync.CurrentPct;
                    if (pct < _lastPct) _distanceFactor = UnityEngine.Random.Range(1 - _distanceRandomJSON.val, 1 + _distanceRandomJSON.val);
                    _lastPct = pct;
                }
                else return;
            }
            if (__DEBUG__) SuperController.LogMessage($"Updating thrust: pct = {pct}, t = {_T(_cycler.CurrentPct, _easingInJSON.val, _easingOutJSON.val)}");

            _penisBaseCtrl.transform.position = _getPos(_T(_cycler.CurrentPct, _easingInJSON.val, _easingOutJSON.val));
        }

        private float _T(float currPct, string easeIn = null, string easeOut = null)
        {
            easeIn = easeIn ?? "Linear";
            easeOut = easeOut ?? "Linear";
            bool r = currPct >= _cyclePctInJSON.val;
            float x = r ? (currPct - _cyclePctInJSON.val) / (1 - _cyclePctInJSON.val) : currPct / _cyclePctInJSON.val;
            return r ? 1 - Easings.EasingFuncs[easeOut](x) : Easings.EasingFuncs[easeIn](x);
        }

        private void Update() { _aligner.Update(); }

        private Vector3 _getPos(float t)
        {
            float u = t * _thrustDistance * _distanceFactor + _distanceOutJSON.val;
            Vector3 basedir = (_targetManager.TargetPos2 - _targetManager.TargetPos1).normalized;
            Quaternion offsetRot = Quaternion.AngleAxis(_dirOffsetXJSON.val, _targetManager.OffsetDirX) * Quaternion.AngleAxis(_dirOffsetYJSON.val, _targetManager.OffsetDirY);
            return offsetRot * basedir * u / 10 + _targetManager.TargetPos1;
        }

        private void _UpdateState()
        {
            if (!_restoring)
            {
                if (_activeJSON.val && enabled && _isValid)
                {
                    if (_thrustJSON.val)
                    {
                        if (_ctrlSnapshot == null) _ctrlSnapshot = new CtrlSnapshot(_penisBaseCtrl);
                        _penisBaseCtrl.RBHoldPositionSpring = 10000;
                        _penisBaseCtrl.currentPositionState = _lockJSON.val ? FreeControllerV3.PositionState.Lock : FreeControllerV3.PositionState.On;
                        _cycler.StartCycle();
                        isActive.val = true;
                    }
                    else if (_cycler.IsActive)
                    {
                        _cycler.StopCycle();
                        isActive.val = false;
                        _RestoreCtrlSnapshot();
                        _ctrlSnapshot = null;
                    }

                    if (!_aligner.IsEnabled && _alignJSON.val) _aligner.Enable();
                    if (_aligner.IsEnabled && !_alignJSON.val) _aligner.Disable();

                }
                else _Deactivate();
            }
        }

        private void _Deactivate()
        {
            _distanceFactor = 1.0f;

            if (_cycler != null && _cycler.IsActive) _cycler.StopCycle();
            isActive.val = false;
            _RestoreCtrlSnapshot();
            _ctrlSnapshot = null;

            if (_aligner.IsEnabled) _aligner.Disable();
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);

            var snapshot = Serializer.Serialize(_ctrlSnapshot);
            if (snapshot != null) json["CtrlSnapshot"] = snapshot;

            var alignerJson = _aligner.GetJSON();
            if (alignerJson != null) json["AlignerData"] = alignerJson;

            needsStore = true;
            return json;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            _restoring = true;
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);

            if (jc.HasKey("CtrlSnapshot")) _ctrlSnapshot = Serializer.DeserializeCtrlSnapshot(jc["CtrlSnapshot"] as JSONClass);
            if (jc.HasKey("AlignerData")) _aligner.RestoreFromJSON(jc["AlignerData"]);

            _restoring = false;
        }

        private void _RestoreCtrlSnapshot()
        {
            if (_ctrlSnapshot == null) return;

            _penisBaseCtrl.currentPositionState = (FreeControllerV3.PositionState)Enum.Parse(typeof(FreeControllerV3.PositionState), _ctrlSnapshot.currentPositionState);
            _penisBaseCtrl.transform.localPosition = _ctrlSnapshot.localPosition;
        }

        private void OnActiveChange(bool _) { _UpdateState(); }
        private void OnThrustEnableChange(bool _) { _UpdateState(); }
        private void OnAlignEnableChange(bool _) { _UpdateState(); }
        private void OnLockPenisChange(bool _) { _UpdateState(); }

        // ========================================== UI ========================================== //

        private void SetupUI()
        {

            _activeJSON = UI_Utils.SetupToggle(this, "<color=#1010A0><size=32><b>Active</b></size></color>", false);
            _activeJSON.setCallbackFunction += OnActiveChange;
            _thrustJSON = UI_Utils.SetupToggle(this, "Enable Thrust", true);
            _thrustJSON.setCallbackFunction += OnThrustEnableChange;
            _alignJSON = UI_Utils.SetupToggle(this, "Enable Alignment", true);
            _alignJSON.setCallbackFunction += OnAlignEnableChange;

            // Target
            _targetManager.SetupUI(this, targetAdjust: false);

            // Settings - Thrust
            CreateSpacer().height = 40f;
            _cycler.SetupUI(this);
            _cyclePctInJSON = UI_Utils.SetupSlider(this, "Duration % In", 0.5f, 0.1f, 0.9f);
            _easingInJSON = new JSONStorableStringChooser("EaseIn", Easings.EasingTypes, "Linear", "Ease In");
            CreateScrollablePopup(_easingInJSON);
            RegisterStringChooser(_easingInJSON);
            _easingOutJSON = new JSONStorableStringChooser("EaseOut", Easings.EasingTypes, "Linear", "Ease Out");
            CreateScrollablePopup(_easingOutJSON);
            RegisterStringChooser(_easingOutJSON);

            _lockJSON = UI_Utils.SetupToggle(this, "Lock Penis Control", false);
            _lockJSON.setCallbackFunction += OnLockPenisChange;

            // Settings - Distances
            CreateSpacer().height = 40f;
            _distanceInJSON = UI_Utils.SetupSlider(this, "Distance In", 0f, -3.0f, 3.0f);
            _distanceOutJSON = UI_Utils.SetupSlider(this, "Distance Out", -2.0f, -3.0f, 3.0f);
            _distanceRandomJSON = UI_Utils.SetupSlider(this, "Distance Random %", 0f, 0f, 0.5f, false);

            // Target adjustments
            CreateSpacer().height = 40f;
            _targetManager.SetupUI(this, false);
            _dirOffsetYJSON = UI_Utils.SetupSlider(this, "<size=24>Angle Adj. Left/Right</size>", 0f, -90f, 90f);
            _dirOffsetXJSON = UI_Utils.SetupSlider(this, "<size=24>Angle Adj. Down/Up</size>", 0f, -90f, 90f);

            CreateSpacer().height = 10f;
            _useMasterJSON = UI_Utils.SetupToggle(this, "Sync with Master", false);
            _useMasterJSON.setCallbackFunction += OnActiveChange;
            _sync.SetupUI(this);
            CreateSpacer().height = 400f;

            // Instructions
            string instructionText =
                "<color=#202020><size=36><b>Settings Information</b></size></color>\n\n" +
                "<color=#101060><b>Activation Settings</b></color>\n" +
                "<b>Enable Thrust</b>: Enables penis thrust. Disable this if you only want penis alignment.\n" +
                "<b>Enable Alignment</b>: When enabled, attempts to point the penis at the target. Disable this if you want to set the penis control rotations manually.\n\n" +

                "<color=#101060><b>Target Settings</b></color>\n" +
                "<b>Target Atom</b>: Name of target Atom.\n" +
                "<b>Target</b>: Anus, Mouth, or Vagina.\n\n" +

                "<color=#101060><b>Cycle Settings</b></color>\n" +
                "<b>Cycle Duration:</b> Duration of a complete <i>In</i> and <i>Out</i> motion.\n" +
                "<b>Cycle Duration % In:</b> Proportion of Cycle Duration taken by the <i>In</i> motion.\n" +
                " 0.5 = <i>In</i> and <i>Out</i> have equal duration.\n" +
                "Less than 0.5 = <i>In</i> motion is faster than <i>Out</i> motion.\n" +
                "More than 0.5 = <i>In</i> motion is slower than <i>Out</i> motion.\n" +
                "<b>Cycle Duration Random %:</b> Proportion of Cycle Duration to randomly add/subtract from each cycle. A value of 0.0 means no random variation, a value of 0.5 means cycles will randomly vary between 50% and 150% of Cycle Duration.\n" +
                "<b>Ease In:</b> Easing function for <i>In</i> motion. See https://easings.net for reference.\n" +
                "<b>Ease Out:</b> Easing function for <i>Out</i> motion.\n" +
                "<b>Lock Penis Control:</b> Sets control state to 'Lock' instead of 'On' to help penis follow motion more accurately. <i>May cause physics instability; use with caution</i>. When enabling this option, it is recommended to set Distance In to a negative value (such as -0.50) and slowly increase it until you get the desired result.\n\n" +

                "<color=#101060><b>Distances</b></color>\n" +
                "<b>Distance In/Out</b>: Endpoints of the thrust motion.  Zero is the target's location. Positive values mean towards the target, and negative values mean away from the target.\n\n" +

                "<color=#101060><b>Adjustments</b></color>\n" +
                "These settings affect both the thrust and alignment.\n" +
                "<b>Target Adj. Left-Right</b>: Offsets the target locations left or right, relative to the target.\n" +
                "<b>Target Adj. Down-Up</b>: Offsets the target locations down or up, relative to the target.\n\n" +

                "<color=#101060><b>Sync Settings</b></color>\n" +
                "When active, use the cycle timing from the specified atom and plugin instance. Cycle Duration and Cycle Duration Random % from this instance will be ignored.\n" +
                "All other settings still apply, such as Cycle Duration % In, easings, and distance settings.\n\n";

            CreateTextField(new JSONStorableString("text", instructionText), true).height = 1200;

        }


    }


}
